# Copyright (c) 2013, Charles O. Goddard and Alethea Butler.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class BitStream:
    def __init__(self, fd):
        self.fd = fd
        self.work_byte = 0
        self.bit_idx = -1
        self.hit_eof = False

    def read_int(self, bits=32, signed=False):
        '''
        Read an n-bit integer from the bitstream.
        '''
        res = 0
        for n in range(bits):
            bit = self.read_bit()
            if bit is None:
                raise EOFError()

            res += bit * 2**n

        # If we are reading a signed number and the most significant
        # bit is set, it's a negative number.
        if signed and res & 1 << (bits-1):
            # Take the two's complement to get the magnitude, then negate.
            res = -((~res + 1) & (2**bits - 1))
        return res

    def read_bit(self):
        '''
        Read a single bit from the bitstream.
        '''
        if self.bit_idx == 8 or self.bit_idx < 0:
            if not self.advance_byte():
                return None

        bit = bool((self.work_byte >> self.bit_idx) & 1)
        self.bit_idx = self.bit_idx + 1
        return bit

    def align(self):
        '''
        Align the stream to a byte boundary.
        '''
        if self.bit_idx != 0 and self.bit_idx != 8:
            self.advance_byte()

    def flush(self):
        '''
        Write out the current byte to the base stream.
        '''
        self.fd.write(bytes([self.work_byte]))

    def write_int(self, n, bits=32, signed=False):
        '''
        Write an n-bit integer.
        '''
        for _ in range(bits):
            self.write_bit(n & 1)
            n = n >> 1

    def write_bit(self, bit):
        '''
        Write a single bit.
        '''
        if self.bit_idx < 0:
            self.advance_byte()

        # Clear the Nth bit in the working byte
        self.work_byte &= 255 ^ (1 << self.bit_idx)
        # Set it to the input bit
        self.work_byte |= bit << self.bit_idx

        self.bit_idx = self.bit_idx + 1
        if self.bit_idx == 8:
            self.fd.write(bytes([self.work_byte]))
            self.advance_byte()

    def advance_byte(self):
        '''
        Advance the stream to the next byte and return True if more
        data exists.
        '''
        self.work_byte = 0
        self.bit_idx = 0
        if not self.hit_eof:
            data = self.fd.read(1)
            if not data:
                self.hit_eof = True
                return False
            else:
                self.work_byte = data[0]
        else:
            return False
        return True
