# Copyright (c) 2013, Charles O. Goddard and Alethea Butler.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from collections import namedtuple


class BitStructure:
    def __init__(self, name, fields):
        self.name = name
        self.fields = fields
        self.tuple = namedtuple(name, ' '.join(f.name for f in fields))

    def deserialize(self, bs):
        values = {}
        for f in self.fields:
            values[f.name] = f.deserialize(bs, values)
        return self.tuple(**values)

    def serialize(self, bs, obj):
        for f in self.fields:
            f.serialize(bs, obj.__dict__, getattr(obj, f.name))

    def __call__(self, *args, **kwargs):
        return self.tuple(*args, **kwargs)


class Type:
    def __init__(self, name):
        self.name = name

    def prepare(self, dict):
        pass

    def deserialize(self, bs, current_state):
        raise NotImplementedError()

    def serialize(self, bs, current_state, value):
        raise NotImplementedError()


class Int(Type):
    def __init__(self, name, bits=32, signed=True, *args, **kwargs):
        self.bits = bits
        self.signed = signed
        super(Int, self).__init__(name, *args, **kwargs)

    def deserialize(self, bs, current_state,):
        return bs.read_int(self.bits, self.signed)

    def serialize(self, bs, current_state, value):
        bs.write_int(value, self.bits, self.signed)


class Size(Int):
    def __init__(self, name, fieldname, signed=False, *args, **kwargs):
        self.fieldname = fieldname
        super(Size, self).__init__(name, signed=signed, *args, **kwargs)

    def serialize(self, bs, current_state, value):
        super(Size, self).serialize(bs, current_state, len(current_state[self.fieldname]))


class Bytes(Type):
    def __init__(self, name, size, padding=0, *args, **kwargs):
        self.size = size
        self.padding = padding
        super(Bytes, self).__init__(name, *args, **kwargs)

    def deserialize(self, bs, current_state):
        res = bytearray()
        if self.size is None:
            while True:
                try:
                    c = bs.read_int(bits=8, signed=False)
                    res.append(c)
                except EOFError:
                    break
        else:
            size = self.size
            if isinstance(self.size, Size):
                size = current_state[size.name]

            for _ in range(size):
                c = bs.read_int(bits=8, signed=False)
                res.append(c)
        return bytes(res)

    def serialize(self, bs, current_state, value):
        if self.size is None or isinstance(self.size, Size):
            for b in value:
                bs.write_int(b, bits=8, signed=False)
        else:
            if len(value) > self.size:
                raise RuntimeError('Would truncate')

            for b in value:
                bs.write_int(b, bits=8, signed=False)
            for _ in range(self.size - len(value)):
                bs.write_int(self.padding, bits=8, signed=False)


class CString(Type):
    def deserialize(self, bs, current_state):
        res = []
        c = 0
        while True:
            c = bs.read_int(bits=8, signed=False)
            if c == 0:
                break
            res.append(chr(c))
        return ''.join(res)

    def serialize(self, bs, current_state, value):
        data = value.encode('latin-1')
        for c in data:
            bs.write_bits(c, bits=8, signed=False)
        bs.write_bits(0, bits=8, signed=False)
